from __future__ import annotations

from contextlib import suppress

import graphene
from aniso8601 import parse_time
from django import forms
from django.db import models  # noqa: TCH002
from graphene_django.converter import convert_django_field, get_django_field_description
from graphene_django.forms.converter import convert_form_field, get_form_field_description
from graphene_django.registry import Registry  # noqa: TCH002
from rest_framework import serializers
from rest_framework.relations import PKOnlyObject

from .converters import convert_typed_dict_to_graphene_type
from .typing import TYPE_CHECKING

if TYPE_CHECKING:
    import datetime
    from typing import Any, TypedDict


__all__ = [
    "IntegerPrimaryKeyField",
    "IntChoiceField",
    "IntMultipleChoiceField",
    "EnumChoiceField",
    "EnumMultipleChoiceField",
    "TypedDictField",
    "TypedDictListField",
]


class IntPkOnlyObject(PKOnlyObject):
    """PK object that is coerced to an integer."""

    def __int__(self) -> int:
        return int(self.pk)


class IntChoiceFieldMixin:
    def __init__(self: forms.TypedChoiceField | forms.TypedMultipleChoiceField, **kwargs: Any) -> None:
        kwargs["coerce"] = int
        kwargs["empty_value"] = None
        super().__init__(**kwargs)

    def valid_value(self: forms.TypedChoiceField | forms.TypedMultipleChoiceField, value: Any) -> bool:
        if self.choices:  # pragma: no cover
            parent: forms.TypedChoiceField = super()  # type: ignore[assignment]
            return parent.valid_value(value)
        try:
            self.coerce(value)
        except (ValueError, TypeError):  # pragma: no cover
            return False
        return True


class EnumChoiceFieldMixin:
    def __init__(self, enum: type[models.Choices], **kwargs: Any) -> None:
        self.enum = enum
        kwargs["choices"] = enum.choices
        super().__init__(**kwargs)


class EnumChoiceFilterMixin:
    def __init__(self, enum: type[models.Choices], *args: Any, **kwargs: Any) -> None:
        kwargs["enum"] = enum
        kwargs["choices"] = enum.choices
        super().__init__(*args, **kwargs)


class TypedDictFieldMixin:
    def __init__(self, typed_dict: type[TypedDict], *arg: Any, **kwargs: Any) -> None:  # pragma: no cover
        type_ = convert_typed_dict_to_graphene_type(typed_dict)
        super().__init__(type_, *arg, **kwargs)


class IntegerPrimaryKeyField(serializers.PrimaryKeyRelatedField, serializers.IntegerField):
    """A field that refers to foreign keys by an integer primary key."""

    def get_attribute(self, instance: models.Model) -> IntPkOnlyObject | None:
        attribute = super().get_attribute(instance)
        if isinstance(attribute, PKOnlyObject) and attribute.pk:
            return IntPkOnlyObject(pk=attribute.pk)
        return None  # pragma: no cover


class IntChoiceField(IntChoiceFieldMixin, forms.TypedChoiceField):
    """Allow plain integers as choices in GraphQL filters. Supports a single choice."""


class IntMultipleChoiceField(IntChoiceFieldMixin, forms.TypedMultipleChoiceField):
    """Same as `IntChoiceField` above but supports multiple choices."""


class EnumChoiceField(EnumChoiceFieldMixin, forms.ChoiceField):
    """
    Custom field for handling enums better in GraphQL filters. Supports a single choice.
    See `EnumChoiceFilter` for motivation.
    """


class EnumMultipleChoiceField(EnumChoiceFieldMixin, forms.MultipleChoiceField):
    """Same as `EnumChoiceField` but supports multiple choices."""


class TypedDictField(TypedDictFieldMixin, graphene.Field):
    """Field that converts a TypedDict to a graphene Field ObjectType."""


class TypedDictListField(TypedDictFieldMixin, graphene.List):
    """Field that converts a TypedDict to a graphene List ObjectType."""


class Time(graphene.Time):
    """Time scalar that can parse time-strings from database."""

    @staticmethod
    def serialize(time: datetime.time | str) -> str:
        if isinstance(time, str):
            with suppress(ValueError):
                time = parse_time(time)
        return graphene.Time.serialize(time)


@convert_django_field.register
def convert_time_to_string(field: models.TimeField, registry: Registry | None = None) -> Time:
    return Time(description=get_django_field_description(field), required=not field.null)


@convert_form_field.register
def convert_form_field_to_time(field: forms.TimeField) -> Time:
    return Time(description=get_form_field_description(field), required=field.required)


@convert_form_field.register
def convert_form_field_to_int(field: IntChoiceField) -> graphene.Int:
    return graphene.Int(description=get_form_field_description(field), required=field.required)


@convert_form_field.register
def convert_form_field_to_list_of_int(field: IntMultipleChoiceField) -> graphene.List:
    return graphene.List(graphene.Int, description=get_form_field_description(field), required=field.required)


@convert_form_field.register
def convert_form_field_to_enum(field: EnumChoiceField) -> graphene.Enum:
    return graphene.Enum.from_enum(field.enum)(
        description=get_form_field_description(field),
        required=field.required,
    )


@convert_form_field.register
def convert_form_field_to_enum_list(field: EnumMultipleChoiceField) -> graphene.List:
    return graphene.List(
        graphene.Enum.from_enum(field.enum),
        description=get_form_field_description(field),
        required=field.required,
    )



from __future__ import annotations

import json
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING

import pytest
import sqlparse
from django import db
from django.contrib.auth import get_user_model
from django.test import Client
from graphene_django.utils.testing import graphql_query

from ..typing import TypedDict

if TYPE_CHECKING:
    from collections.abc import Generator

    from django.contrib.auth.models import User
    from django.http import HttpResponse

    from ..typing import Any, Self


__all__ = [
    "GraphQLClient",
    "capture_database_queries",
]


class FieldError(TypedDict):
    field: str
    messages: list[str]


@dataclass
class QueryData:
    queries: list[str]

    @property
    def log(self) -> str:
        message = "-" * 75
        message += f"\n>>> Queries ({len(self.queries)}):\n"
        for index, query in enumerate(self.queries):
            message += f"{index + 1}) ".ljust(75, "-") + f"\n{query}\n"
        message += "-" * 75
        return message


class GQLResponse:
    def __init__(self, response: HttpResponse, query_data: QueryData) -> None:
        # 'django.test.client.Client.request' sets json attribute on the response.
        self.json: dict[str, Any] = response.json()  # type: ignore[attr-defined]
        self.query_data = query_data

    def __str__(self) -> str:
        return json.dumps(self.json, indent=2, sort_keys=True, default=str)

    def __repr__(self) -> str:
        return repr(self.json)

    def __len__(self) -> int:
        return len(self.edges)

    def __getitem__(self, item: str) -> dict[str, Any] | None:
        return (self.data or {})[item]

    def __contains__(self, item: str) -> bool:
        return item not in self.json

    @property
    def queries(self) -> list[str]:
        """Return a list of the database queries that were executed."""
        # Note: To use query counting, DEBUG needs to be set to True.
        return self.query_data.queries

    @property
    def query_log(self) -> str:
        """Return a string representation of the database queries that were executed."""
        # Note: To use query counting, DEBUG needs to be set to True.
        return self.query_data.log

    @property
    def data(self) -> dict[str, Any] | None:
        """Return the data from the response content."""
        return self.json["data"]

    @property
    def first_query_object(self) -> dict[str, Any] | None:
        """
        Return the first query object in the response content.

        >>> self.json = {"data": {"foo": {"name": "bar"}}}
        >>> self.first_query_object
        {"name": "bar"}
        """
        data = self.data or {}
        try:
            return next(iter(data.values()))
        except StopIteration:  # pragma: no cover
            pytest.fail(f"No query object not found in response content: {self.json}")

    @property
    def edges(self) -> list[dict[str, Any]]:
        """
        Return edges from the first query in the response content.

        >>> self.json = {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.edges
        [{"node": {"name": "bar"}}]
        """
        try:
            return self.first_query_object["edges"]
        except (KeyError, TypeError):  # pragma: no cover
            pytest.fail(f"Edges not found in response content: {self.json}")

    def node(self, index: int = 0) -> dict[str, Any]:
        """
        Return the node at the given index in the response content edges.

        >>> self.json = {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.node(0)
        {"name": "bar"}
        """
        try:
            return self.edges[index]["node"]
        except (IndexError, TypeError):  # pragma: no cover
            pytest.fail(f"Node {index!r} not found in response content: {self.json}")

    @property
    def has_errors(self) -> bool:  # pragma: no cover
        """Are there any errors in the response?"""
        # Errors in the root of the response
        if "errors" in self.json and self.json.get("errors") is not None:
            return True

        # Errors in the fields of the first query object
        if "errors" in (self.first_query_object or {}) and (self.first_query_object or {}).get("errors") is not None:
            return True

        return False

    @property
    def errors(self) -> list[dict[str, Any]]:
        """
        Return errors found in the root of the response.

        >>> self.json = {"errors": [{"locations": [...], "message": "bar", "path": [...]}]}
        >>> self.errors
        [{"locations": [...], "message": "bar", "path": [...]}]
        """
        try:
            return self.json["errors"]
        except (KeyError, TypeError):  # pragma: no cover
            pytest.fail(f"Errors not found in response content: {self.json}")

    def error_message(self, selector: int | str = 0) -> str:  # pragma: no cover
        """
        Return the error message from the errors list...

        1) in the given index

        >>> self.json = {"errors": [{"message": ["bar"], "path": [...]}]}
        >>> self.error_message(0)  # default
        "bar"

        2) in the given path:

        >>> self.json = {"errors": [{"message": ["bar"], "path": ["fizz", "buzz", "foo"]}]}
        >>> self.error_message("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["message"]
            except IndexError:
                pytest.fail(f"Errors list doesn't have an index {selector}: {self.json}")
            except (KeyError, TypeError):
                pytest.fail(f"Field 'message' not found in error content: {self.json}")
        else:
            try:
                return next(error["message"] for error in self.errors if error["path"][-1] == selector)
            except StopIteration:
                pytest.fail(f"Errors list doesn't have an error for field '{selector}': {self.json}")
            except (KeyError, TypeError):
                pytest.fail(f"Field 'message' not found in error content: {self.json}")

    @property
    def field_errors(self) -> list[FieldError]:
        """
        Return errors for data fields.

        >>> self.json = {"data": {"foo": {"errors": [{"field": "bar", "message": ["baz"]}]
        >>> self.field_errors
        [{"field": "bar", "message": ["baz"]}]
        """
        try:
            return self.first_query_object["errors"]
        except (KeyError, TypeError):  # pragma: no cover
            pytest.fail(f"Field errors not found in response content: {self.json}")

    def field_error_messages(self, field: str = "nonFieldErrors") -> list[str]:  # pragma: no cover
        """
        Return field error messages for desired field.

        >>> self.json = {"errors": [{"field": "foo", "message": ["bar"]}]}
        >>> self.field_error_messages("foo")
        ["bar"]
        """
        for error in self.field_errors or []:
            if error.get("field") == field:
                try:
                    return error["messages"]
                except (KeyError, TypeError):
                    msg = f"Error message for field {field!r} not found in error: {error}"
                    pytest.fail(msg)

        msg = f"Error for field {field!r} not found in response content: {self.json}"
        raise pytest.fail(msg)

    def assert_query_count(self, count: int) -> None:  # pragma: no cover
        # Note: To use query counting, DEBUG needs to be set to True.
        assert len(self.queries) == count, self.query_log  # noqa: S101


class GraphQLClient(Client):
    def __call__(  # noqa: PLR0913
        self: Self,
        query: str,
        operation_name: str | None = None,
        input_data: dict[str, Any] | None = None,
        variables: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
    ) -> GQLResponse:
        """
        Make a GraphQL query to the test client.

        :params query: GraphQL query string.
        :params operation_name: Name of the operation to execute.
        :params input_data: Set (and override) the "input" variable in the given variables.
        :params variables: Variables for the query.
        :params headers: Headers for the query. Keys should in all-caps, and be prepended with "HTTP_".
        """
        with capture_database_queries() as results:
            response = graphql_query(
                query=query,
                operation_name=operation_name,
                input_data=input_data,
                variables=variables,
                headers=headers,
                client=self,
            )
        return GQLResponse(response, results)

    def login_with_superuser(self) -> User:
        user = get_user_model().objects.create_superuser(username="superuser", email="superuser@django.com")
        self.force_login(user)
        return user

    def login_with_regular_user(self) -> User:
        user = get_user_model().objects.create_user(username="user", email="user@django.com")
        self.force_login(user)
        return user


@contextmanager
def capture_database_queries() -> Generator[QueryData, None, None]:
    """Capture results of what database queries were executed. `DEBUG` needs to be set to True."""
    results = QueryData(queries=[])
    db.connection.queries_log.clear()

    try:
        yield results
    finally:
        results.queries = [
            sqlparse.format(query["sql"], reindent=True)
            for query in db.connection.queries
            if "sql" in query
            and not query["sql"].startswith("SAVEPOINT")
            and not query["sql"].startswith("RELEASE SAVEPOINT")
        ]



from __future__ import annotations

import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Literal,
    Mapping,
    NamedTuple,
    ParamSpec,
    Protocol,
    Sequence,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
)

# New in version 3.11
if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:  # pragma: no cover
    from typing import Self

from django.core.handlers.wsgi import WSGIRequest
from graphql import GraphQLResolveInfo

if TYPE_CHECKING:
    import django_filters
    from django.contrib.auth.models import AnonymousUser, User
    from django.db.models import Field, Model, QuerySet
    from django.forms import Form
    from graphql_relay import Edge, PageInfo
    from rest_framework.serializers import ListSerializer

    from .bases import DjangoNode
    from .serializers import NestingModelSerializer


__all__ = [
    "Any",
    "AnyUser",
    "Callable",
    "FieldNameStr",
    "FilterOverride",
    "FilterSetMeta",
    "Generic",
    "GQLInfo",
    "Literal",
    "Mapping",
    "NamedTuple",
    "NodeConnectionType",
    "ParamSpec",
    "PermCheck",
    "Protocol",
    "RelatedSerializer",
    "RelationType",
    "Self",
    "Sequence",
    "SerializerMeta",
    "TypeAlias",
    "TypedDict",
    "TypeVar",
    "Union",
]


FieldNameStr: TypeAlias = str
FieldLookupStr: TypeAlias = str
FilterAliasStr: TypeAlias = str
LookupNameStr: TypeAlias = str
MethodNameStr: TypeAlias = str
AnyUser: TypeAlias = Union["User", "AnonymousUser"]
PermCheck: TypeAlias = Callable[[AnyUser], bool] | Callable[[AnyUser, "Model"], bool]
RelationType: TypeAlias = Literal["one_to_one", "many_to_one", "one_to_many", "many_to_many"]
RelatedSerializer: TypeAlias = Union["NestingModelSerializer", "ListSerializer"]

T = TypeVar("T")


class NodeConnectionType(Protocol[T]):
    page_info: PageInfo
    edged: list[Edge]
    iterable: QuerySet[T]
    length: int


class OrderingFunc(Protocol):
    def __call__(self, qs: QuerySet, *, desc: bool) -> QuerySet:
        """Custom ordering function."""


class FilterOverride(TypedDict):
    filter_class: type[django_filters.Filter]
    extra: Callable[[Field], dict[str, Any]]


class UserHintedWSGIRequest(WSGIRequest):
    user: AnyUser


class GQLInfo(GraphQLResolveInfo):
    context = UserHintedWSGIRequest


class SerializerMeta:
    model: type[Model]
    fields: Sequence[FieldNameStr] | Literal["__all__"]
    read_only_fields: Sequence[FieldNameStr]
    exclude: Sequence[FieldNameStr]
    depth: int
    extra_kwargs: Mapping[FieldNameStr, Mapping[str, Any]]
    node: DjangoNode


class FilterSetMeta:
    model: type[Model] | None
    fields: Sequence[FieldNameStr] | Mapping[FieldNameStr, Sequence[LookupNameStr]] | Literal["__all__"] | None
    exclude: Sequence[FieldNameStr] | None
    filter_overrides: Mapping[Field, FilterOverride]
    form: type[Form]
    combination_methods: Sequence[MethodNameStr]
    order_by: Sequence[FieldLookupStr | tuple[FieldLookupStr, FilterAliasStr]]



from .bases import CreateMutation, DeleteMutation, DjangoNode, UpdateMutation
from .filters import ModelFilterSet
from .serializers import NestingModelSerializer

__all__ = [
    "DjangoNode",
    "CreateMutation",
    "UpdateMutation",
    "DeleteMutation",
    "NestingModelSerializer",
    "ModelFilterSet",
]



from __future__ import annotations

from enum import Enum
from functools import partial
from typing import TYPE_CHECKING

import graphene
from django.core.exceptions import ValidationError as DjangoValidationError
from graphene import ClientIDMutation, Field, InputField
from graphene.types.resolver import attr_resolver
from graphene.types.utils import yank_fields_from_attrs
from graphene_django import DjangoConnectionField, DjangoListField, DjangoObjectType
from graphene_django.converter import convert_django_field
from graphene_django.filter import DjangoFilterConnectionField
from graphene_django.registry import get_global_registry
from graphene_django.rest_framework.mutation import fields_for_serializer
from graphene_django.settings import graphene_settings
from graphene_django.types import ALL_FIELDS, ErrorType
from graphene_django.utils import camelize, is_valid_django_model
from graphql_relay import to_global_id
from rest_framework.exceptions import ValidationError as SerializerValidationError
from rest_framework.fields import SerializerMethodField
from rest_framework.serializers import ListSerializer, ModelSerializer, as_serializer_error
from rest_framework.settings import api_settings

from .connections import Connection
from .converters import convert_serializer_fields_to_not_required
from .errors import flatten_errors
from .model_operations import get_model_lookup_field, get_object_or_404
from .options import DjangoMutationOptions, DjangoNodeOptions
from .permissions import AllowAny, BasePermission, restricted_field
from .settings import gdx_settings
from .typing import Sequence
from .utils import get_filters_from_info

if TYPE_CHECKING:
    from django.db import models
    from django.http import HttpRequest
    from graphene.relay.node import NodeField
    from graphene.types.mountedtype import MountedType
    from graphene.types.unmountedtype import UnmountedType

    from .typing import Any, AnyUser, FieldNameStr, GQLInfo, Literal, PermCheck, Self


__all__ = [
    "DjangoNode",
    "CreateMutation",
    "UpdateMutation",
    "DeleteMutation",
]


class DjangoNode(DjangoObjectType):
    """
    Custom base class for GraphQL-types that are backed by a Django model.
    Adds the following features to all types that inherit it:

    - Makes the `graphene.relay.Node` interface the default interface for the type.

    - Adds the `errors` list-field to the type for returning errors.

    - Adds the `totalCount` field the default Connection field for the type.

    - Adds the `filter_queryset` method that can be overridden to add additional filtering for both
      `get_queryset` and `get_node` query sets.

    - Adds convenience methods for creating fields/list-fields/nodes/connections for the type.

    - Adds the `pk` field and resolver to the type if present on Meta.fields.

    ---

    The following options can be set in the `Meta`-class.

    `model: type[Model]`

    - Required. The model class for the node.

    `fields: list[FieldNameStr] | Literal["__all__"]`

    - Required. The fields to include in the node. If `__all__` is used, all fields are included.

    `permission_classes: Sequence[type[BasePermission]]`

    - Optional. Set permission classes for the node. Defaults to (`AllowAny`,).

    `restricted_fields: dict[FieldNameStr, PermCheck]`

    - Optional. Adds permission checks to the resolvers of the fields as defined in the dict.
    """

    _meta: DjangoNodeOptions

    class Meta:
        abstract = True

    errors = graphene.List(ErrorType, description="May contain more than one error for same field.")

    @classmethod
    def Field(cls, **kwargs: Any) -> Field:  # noqa: N802
        return Field(cls, **kwargs)

    @classmethod
    def ListField(cls, **kwargs: Any) -> DjangoListField:  # noqa: N802
        return DjangoListField(cls, **kwargs)

    @classmethod
    def Node(cls, **kwargs: Any) -> NodeField:  # noqa: N802
        return graphene.relay.Node.Field(cls, **kwargs)

    @classmethod
    def Connection(cls, **kwargs: Any) -> DjangoFilterConnectionField | DjangoConnectionField:  # noqa: N802
        if cls._meta.filterset_class is None and cls._meta.filter_fields is None:
            return DjangoConnectionField(cls, **kwargs)  # pragma: no cover
        return DjangoFilterConnectionField(cls, **kwargs)

    @classmethod
    def filter_queryset(cls, queryset: models.QuerySet, info: GQLInfo) -> models.QuerySet:
        """Implement this method filter to the available rows from the model on this node."""
        return queryset

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: type[models.Model] | None = None,
        fields: list[FieldNameStr] | Literal["__all__"] | None = None,
        permission_classes: Sequence[type[BasePermission]] = (AllowAny,),
        restricted_fields: dict[FieldNameStr, PermCheck] | None = None,
        **options: Any,
    ) -> None:
        if model is None:  # pragma: no cover
            msg = "`Meta.model` is required."
            raise ValueError(msg)

        if not is_valid_django_model(model):  # pragma: no cover
            msg = f"`Meta.model` needs to be a Model-class, received: `{model}`."
            raise TypeError(msg)

        if fields is None:  # pragma: no cover
            msg = "`Meta.fields` is required."
            raise ValueError(msg)

        if not (fields == ALL_FIELDS or isinstance(fields, Sequence)):  # pragma: no cover
            msg = f"`Meta.fields` is should be a Sequence of field names or `{ALL_FIELDS}`, received: `{fields}`."
            raise TypeError(msg)

        if not hasattr(cls, "pk") and (fields == ALL_FIELDS or "pk" in fields):
            cls._add_pk_field(model)

        if restricted_fields is not None:
            cls._add_field_restrictions(fields, restricted_fields)

        _meta = DjangoNodeOptions(class_type=cls, permission_classes=permission_classes)
        options.setdefault("connection_class", Connection)
        options.setdefault("interfaces", (graphene.relay.Node,))

        super().__init_subclass_with_meta__(_meta=_meta, model=model, fields=fields, **options)

    @classmethod
    def _add_pk_field(cls, model: type[models.Model]) -> None:
        cls.pk = graphene.Int() if model._meta.pk.name == "id" else graphene.ID()
        cls.resolve_pk = cls.resolve_id

    @classmethod
    def _add_field_restrictions(
        cls,
        fields: Sequence[FieldNameStr] | Literal["__all__"],
        restricted_fields: dict[FieldNameStr, PermCheck],
    ) -> None:
        for field_name, check in restricted_fields.items():
            if fields != ALL_FIELDS and field_name not in fields:  # pragma: no cover
                msg = f"Field `{field_name}` not in `Meta.fields`."
                raise ValueError(msg)

            resolver = getattr(cls, f"resolve_{field_name}", None)
            if resolver is None:
                resolver = partial(attr_resolver, field_name, None)  # must be positional args!

            setattr(cls, f"resolve_{field_name}", restricted_field(check)(resolver))

    @classmethod
    def get_queryset(cls, queryset: models.QuerySet, info: GQLInfo) -> models.QuerySet:
        """Override `filter_queryset` instead of this method to add filtering of possible rows."""
        if not cls.has_filter_permissions(info):
            msg = gdx_settings.FILTER_PERMISSION_ERROR_MESSAGE
            raise PermissionError(msg)
        return cls.filter_queryset(queryset, info)

    @classmethod
    def get_node(cls, info: GQLInfo, pk: Any) -> models.Model | None:
        """Override `filter_queryset` instead of this method to add filtering of possible rows."""
        if not cls.has_node_permissions(info, pk):
            msg = gdx_settings.QUERY_PERMISSION_ERROR_MESSAGE
            raise PermissionError(msg)
        queryset = cls._meta.model.objects.filter(pk=pk)
        return cls.filter_queryset(queryset, info).first()

    @classmethod
    def has_filter_permissions(cls, info: GQLInfo) -> bool:
        """Check which permissions are required to access lists of this type."""
        filters = get_filters_from_info(info)
        return all(perm.has_filter_permission(info.context.user, filters) for perm in cls._meta.permission_classes)

    @classmethod
    def has_node_permissions(cls, info: GQLInfo, pk: Any) -> bool:
        """Check which permissions are required to access single items of this type."""
        filters = get_filters_from_info(info)
        return all(perm.has_node_permission(info.context.user, pk, filters) for perm in cls._meta.permission_classes)

    @classmethod
    def get_global_id(cls, pk: Any) -> str:
        """Get id used for `node` queries."""
        return to_global_id(cls.__name__, pk)


class DjangoMutation(ClientIDMutation):
    """
    Custom base class for GraphQL-mutations that are backed by a Django model.
    Adds the following features to all types that inherit it:

    - Adds the `errors` list-field to the type for returning errors.

    - For updates, converts all fields to optional fields, enabling partial updates.

    - Checks for missing object types for nested model serializer fields to avoid nebulous import order errors.

    - Can add permission checks via permission classes.

    - Formats errors raised from nested serializers to a compatible form.

    ---

    The following options can be set in the `Meta`-class.

    `model: type[models.Model]`

    - Required (delete only). Model class for the model the operation is performed on.

    `serializer_class: type[ModelSerializer]`

    - Required (create and update only). The serializer used for the mutation.

    `output_serializer_class: type[ModelSerializer]`

    - Optional. The serializer used for the output data. If not set, `serializer_class` is used.
      Serializer fields are modified so that all fields are optional, enabling partial updates.

    `permission_classes: Sequence[type[BasePermission]]`

    - Optional. Set permission classes for the mutation. Defaults to (`AllowAny`,).

    `lookup_field: str`

    - Optional. The field used for looking up the instance to be mutated. Defaults to the object's
      primary key, which is usually `id`. Note that the `lookup_field` attribute has to be available
      from the serializer's `Meta.fields` definition.
    """

    _meta: DjangoMutationOptions

    errors = graphene.List(ErrorType, description="May contain more than one error for same field.")

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(  # noqa: PLR0913
        cls,
        lookup_field: FieldNameStr | None = None,
        model: type[models.Model] | None = None,
        serializer_class: type[ModelSerializer] | None = None,
        output_serializer_class: type[ModelSerializer] | None = None,
        permission_classes: Sequence[type[BasePermission]] = (AllowAny,),
        model_operation: Literal["create", "update", "delete", "custom"] = "custom",
        input_fields: dict[str, MountedType | UnmountedType] | None = None,
        output_fields: dict[str, MountedType | UnmountedType] | None = None,
        **options: Any,
    ) -> None:
        if model_operation == "delete":
            if model is None:  # pragma: no cover
                msg = "`Meta.model` is required."
                raise ValueError(msg)

            lookup_field = get_model_lookup_field(model, lookup_field)
            input_field = graphene.ID(required=True)
            if lookup_field != "pk":  # pragma: no cover
                input_field = convert_django_field(model._meta.get_field(lookup_field))

            input_fields = {lookup_field: input_field}
            output_fields = {"deleted": graphene.Boolean()}

        elif model_operation in ("create", "update"):
            if serializer_class is None:  # pragma: no cover
                msg = "`Meta.serializer_class` is required for create and update mutations."
                raise ValueError(msg)

            if not issubclass(serializer_class, ModelSerializer):  # pragma: no cover
                msg = "`Meta.serializer_class` needs to be a ModelSerializer."
                raise TypeError(msg)

            if model_operation == "update":
                serializer_class = convert_serializer_fields_to_not_required(serializer_class, lookup_field)

            serializer = output_serializer = serializer_class()

            if output_serializer_class is not None:  # pragma: no cover
                output_serializer = output_serializer_class()
            else:
                output_serializer_class = serializer_class

            _check_serializer_field_models_in_registry(cls, output_serializer)

            model: type[models.Model] = serializer_class.Meta.model
            lookup_field = get_model_lookup_field(model, lookup_field)

            input_fields = fields_for_serializer(
                serializer,
                only_fields=(),
                exclude_fields=(),
                is_input=True,
                lookup_field=lookup_field,
            )

            output_fields = fields_for_serializer(
                output_serializer,
                only_fields=(),
                exclude_fields=(),
                lookup_field=lookup_field,
            )

        elif cls.custom_model_operation == DjangoMutation.custom_model_operation:  # pragma: no cover
            msg = "`custom_model_operation` must be overridden in subclasses for custom model operations."
            raise ValueError(msg)

        input_fields = yank_fields_from_attrs(attrs=input_fields or {}, _as=InputField)
        output_fields = yank_fields_from_attrs(attrs=output_fields or {}, _as=Field)

        _meta = DjangoMutationOptions(
            class_type=cls,
            model_class=model,
            model_operation=model_operation,
            lookup_field=lookup_field,
            fields=output_fields,
            serializer_class=serializer_class,
            output_serializer_class=output_serializer_class,
            permission_classes=permission_classes,
        )

        super().__init_subclass_with_meta__(_meta=_meta, input_fields=input_fields, **options)

    @classmethod
    def has_permission(cls, root: Any, info: GQLInfo, input_data: dict[str, Any]) -> bool:
        filters = get_filters_from_info(info)
        user = info.context.user
        perms = cls._meta.permission_classes

        if cls._meta.model_operation == "create":
            return all(perm.has_create_permission(root, user, input_data, filters) for perm in perms)
        if cls._meta.model_operation == "update":
            return all(perm.has_update_permission(root, user, input_data, filters) for perm in perms)
        if cls._meta.model_operation == "delete":
            return all(perm.has_delete_permission(root, user, input_data, filters) for perm in perms)
        return all(perm.has_mutation_permission(root, user, input_data, filters) for perm in perms)

    @classmethod
    def mutate(cls, root: Any, info: GQLInfo, **kwargs: Any) -> Self:
        if not cls.has_permission(root=root, info=info, input_data=kwargs["input"]):
            detail = {api_settings.NON_FIELD_ERRORS_KEY: [gdx_settings.MUTATION_PERMISSION_ERROR_MESSAGE]}
            errors = ErrorType.from_errors(detail)
            return cls(errors=errors)  # type: ignore[arg-type]

        try:
            return super().mutate(root=root, info=info, input=kwargs["input"])
        except (DjangoValidationError, SerializerValidationError) as error:
            detail = as_serializer_error(error)
            detail = camelize(detail) if graphene_settings.CAMELCASE_ERRORS else detail
            detail = flatten_errors(detail)
            errors = [ErrorType(field=key, messages=value) for key, value in detail.items()]  # type: ignore[arg-type]
            return cls(errors=errors)  # type: ignore[arg-type]

    @classmethod
    def mutate_and_get_payload(cls, root: Any, info: GQLInfo, **kwargs: Any) -> Self:
        if cls._meta.model_operation == "delete":
            return cls.delete(root=root, info=info, **kwargs)
        if cls._meta.model_operation in ("create", "update"):
            return cls.update_or_delete(root=root, info=info, **kwargs)
        return cls.custom_model_operation(root=root, info=info, **kwargs)

    @classmethod
    def custom_model_operation(cls, root: Any, info: GQLInfo, **kwargs: Any) -> Self:  # pragma: no cover
        """
        Override this method to perform custom model operations in this mutation
        instead of the regular create, update or delete.
        """
        raise NotImplementedError(f"Custom model operation not defined for '{cls.__name__}'")  # noqa: EM102

    @classmethod
    def get_instance(cls, **kwargs: Any) -> models.Model:
        lookup_field: str = cls._meta.lookup_field
        if lookup_field not in kwargs:  # pragma: no cover
            raise SerializerValidationError({lookup_field: "This field is required."})
        return get_object_or_404(cls._meta.model_class, **{lookup_field: kwargs[lookup_field]})

    @classmethod
    def update_or_delete(cls, root: Any, info: GQLInfo, **kwargs: Any) -> Self:
        kwargs = cls.get_serializer_kwargs(info.context, **kwargs)
        serializer = cls._meta.serializer_class(**kwargs)
        serializer.is_valid(raise_exception=True)
        obj = serializer.save()
        output = cls.to_representation(obj)
        return cls(errors=None, **output)  # type: ignore[arg-type]

    @classmethod
    def get_serializer_kwargs(cls, request: HttpRequest, **kwargs: Any) -> dict[str, Any]:
        for input_dict_key, maybe_enum in kwargs.items():
            if isinstance(maybe_enum, Enum):
                kwargs[input_dict_key] = maybe_enum.value

        instance: models.Model | None = None
        if cls._meta.model_operation == "update":
            instance = cls.get_instance(**kwargs)

        return {
            "instance": instance,
            "data": kwargs,
            "context": {"request": request},
            "partial": instance is not None,
        }

    @classmethod
    def to_representation(cls, obj: models.Model) -> dict[str, Any]:
        serializer = cls._meta.output_serializer_class(instance=obj)

        kwargs: dict[str, Any] = {}
        for field_name, field in serializer.fields.items():
            if not field.write_only:
                if isinstance(field, SerializerMethodField):  # pragma: no cover
                    kwargs[field_name] = field.to_representation(obj)
                else:
                    kwargs[field_name] = field.get_attribute(obj)
        return kwargs

    @classmethod
    def delete(cls, root: Any, info: GQLInfo, **kwargs: Any) -> Self:
        instance = cls.get_instance(**kwargs)
        cls.validate_deletion(instance, info.context.user)
        count, _ = instance.delete()
        return cls(errors=None, deleted=bool(count))  # type: ignore[arg-type]

    @classmethod
    def validate_deletion(cls, instance: models.Model, user: AnyUser) -> None:
        """Implement to perform additional validation before the given instance is deleted."""
        return  # pragma: no cover


class CreateMutation(DjangoMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        super().__init_subclass_with_meta__(model_operation="create", **options)


class UpdateMutation(DjangoMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        super().__init_subclass_with_meta__(model_operation="update", **options)


class DeleteMutation(DjangoMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        super().__init_subclass_with_meta__(model_operation="delete", **options)


def _check_serializer_field_models_in_registry(
    mutation_class: type[DjangoMutation],
    serializer: ModelSerializer,
) -> None:
    global_registry = get_global_registry()

    for field in serializer.fields.values():
        if isinstance(field, ModelSerializer):
            field_model: type[models.Model] = field.Meta.model
        elif isinstance(field, ListSerializer) and isinstance(field.child, ModelSerializer):
            field = field.child  # noqa: PLW2901
            field_model: type[models.Model] = field.Meta.model
        else:
            continue

        type_ = global_registry.get_type_for_model(field_model)
        if type_ is None:  # pragma: no cover
            # This error is in place since the mutation class needs to know the ObjectTypes
            # of the models of nested model serializer fields when converting them to ObjectTypes,
            # but currently does not properly warn against this error itself. See this line in
            # `graphene_django.rest_framework.serializer_converter.convert_serializer_field`:
            #
            # args = [global_registry.get_type_for_model(field_model)]
            #
            # Since model is not registered, `global_registry.get_type_for_model` returns `None`,
            # which will cause errors later on in the mutation class creation process.
            msg = (
                f"Could not find a ObjectType for model: `{field_model.__name__}`. "
                f"Make sure that the ObjectType for this model is registered before the "
                f"`{mutation_class.__name__}` mutation class is crated. This can be archived by, "
                f"for example, setting `{field.__class__.__name__}.Meta.node` to that ObjectType."
            )
            raise LookupError(msg)

        _check_serializer_field_models_in_registry(mutation_class, field)



from __future__ import annotations

from typing import TYPE_CHECKING

from graphql.language.ast import (
    BooleanValueNode,
    ConstListValueNode,
    EnumValueNode,
    FieldNode,
    FloatValueNode,
    IntValueNode,
    ListValueNode,
    StringValueNode,
    ValueNode,
    VariableNode,
)

if TYPE_CHECKING:
    from .typing import Any, GQLInfo


__all__ = [
    "get_nested",
    "get_filters_from_info",
]


def get_nested(obj: dict | list | None, /, *args: str | int, default: Any = None) -> Any:
    """
    Get value from a nested structure containing dicts with string keys or lists,
    where the keys and list indices might not exist.

    1) `data["foo"][0]["bar"]["baz"]`
     - Might raise a `KeyError` or `IndexError` if any of the keys or indices don't exist.

    2) `get_nested(data, "foo", 0, "bar", "baz")`
     - Will return `None` (default) if any of the keys or indices don't exist.
    """
    if not args:
        return obj if obj is not None else default

    arg, args = args[0], args[1:]

    if isinstance(arg, int):
        obj = obj or []
        try:
            obj = obj[arg]
        except IndexError:
            obj = None
        return get_nested(obj, *args, default=default)

    obj = obj or {}
    return get_nested(obj.get(arg), *args, default=default)


def get_filters_from_info(info: GQLInfo) -> dict[str, Any]:
    """Find filter arguments in the GraphQL query and return them as a dict."""
    filters: dict[str, Any] = {}

    for field in info.operation.selection_set.selections:
        if not isinstance(field, FieldNode):  # pragma: no cover
            continue
        for argument in field.arguments:
            filters[argument.name.value] = _get_filter_argument(argument.value, info.variable_values)

    return filters


def _get_filter_argument(value: ValueNode, variable_values: dict[str, Any]) -> Any:
    if isinstance(value, (IntValueNode, FloatValueNode, StringValueNode, BooleanValueNode, EnumValueNode)):
        return value.value
    if isinstance(value, (ListValueNode, ConstListValueNode)):
        return [_get_filter_argument(val, variable_values) for val in value.values]
    if isinstance(value, VariableNode):  # pragma: no cover
        return variable_values[value.name.value]

    msg = f"Unsupported ValueNode for filter argument: '{type(value).__name__}'"  # pragma: no cover
    raise ValueError(msg)  # pragma: no cover



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from .client import GraphQLClient, QueryData, capture_database_queries

if TYPE_CHECKING:
    from contextlib import AbstractContextManager

    from pytest_django.fixtures import SettingsWrapper

    from ..typing import Callable


@pytest.fixture()
def graphql() -> GraphQLClient:  # pragma: no cover
    return GraphQLClient()


@pytest.fixture()
def query_counter(settings: SettingsWrapper) -> Callable[[], AbstractContextManager[QueryData]]:  # pragma: no cover
    settings.DEBUG = True
    return capture_database_queries



from __future__ import annotations

import datetime
from copy import deepcopy
from typing import TYPE_CHECKING

import graphene
from django.db import models
from graphene_django.converter import convert_django_field
from rest_framework.serializers import ListSerializer, ModelSerializer

from .model_operations import get_model_lookup_field

if TYPE_CHECKING:
    from graphene.types.unmountedtype import UnmountedType

    from .typing import FieldNameStr, SerializerMeta, TypedDict


__all__ = [
    "convert_typed_dict_to_graphene_type",
    "convert_serializer_fields_to_not_required",
]


_CONVERSION_TABLE: dict[type, type[models.Field]] = {
    int: models.IntegerField,
    str: models.CharField,
    bool: models.BooleanField,
    float: models.FloatField,
    dict: models.JSONField,
    list: models.JSONField,
    set: models.JSONField,
    tuple: models.JSONField,
    bytes: models.BinaryField,
    datetime.datetime: models.DateTimeField,
    datetime.date: models.DateField,
    datetime.time: models.TimeField,
}


def convert_typed_dict_to_graphene_type(typed_dict: type[TypedDict]) -> type[graphene.ObjectType]:
    graphene_types: dict[str, UnmountedType] = {}
    for field_name, type_ in typed_dict.__annotations__.items():
        model_field = _CONVERSION_TABLE.get(type_)
        if model_field is None:
            msg = f"Cannot convert field `{field_name}` of type `{type_.__name__}` to model field."
            raise ValueError(msg)
        graphene_type = convert_django_field(model_field())
        graphene_types[field_name] = graphene_type

    return type(f"{typed_dict.__name__}Type", (graphene.ObjectType,), graphene_types)  # type: ignore[return-value]


def convert_serializer_fields_to_not_required(
    serializer_class: type[ModelSerializer],
    lookup_field: FieldNameStr | None,
    *,
    top_level: bool = True,
) -> type[ModelSerializer | ListSerializer]:
    """
    When updating, usually the wanted behaviour is that the user can only update the
    fields that are actually updated. Therefore, this function can be used to create a new
    serializer class, on which has all the appropriate fields set to `required=False` (except
    the top-level `lookup_field`, which is required to select the row to be updated).
    This function is recursive, so that nested serializers, and their fields are also converted
    to `required=False`.

    :param serializer_class: The serializer class to convert.
    :param lookup_field: The lookup field to be used for the update operation.
    :param top_level: Whether this is the top-level serializer.
    """
    # We need to create a new serializer and rename it since
    # `graphene_django.rest_framework.serializer_converter.convert_serializer_to_input_type`.
    # caches its results by the serializer class name, and thus the input type created for
    # the CREATE operation would be used if created first.
    class_name = f"Update{serializer_class.__name__}"
    new_class: type[ModelSerializer] = type(class_name, (serializer_class,), {})  # type: ignore[assignment]

    # Create a new Meta and deepcopy `extra_kwargs` to avoid changing the original serializer,
    # which might be used for other operations.
    new_class.Meta: SerializerMeta = type("Meta", (serializer_class.Meta,), {})  # type: ignore[assignment]
    new_class.Meta.extra_kwargs = deepcopy(getattr(serializer_class.Meta, "extra_kwargs", {}))

    lookup_field = get_model_lookup_field(new_class.Meta.model, lookup_field)

    for field_name in new_class.Meta.fields:
        _set_extra_kwargs(field_name, new_class.Meta, lookup_field, top_level=top_level)

    # Handle nested serializers
    for field_name, field in new_class._declared_fields.items():
        if isinstance(field, ModelSerializer):
            new_field = convert_serializer_fields_to_not_required(field.__class__, None, top_level=False)
            new_kwargs = field._kwargs | {"required": False}
            new_class._declared_fields[field_name] = new_field(*field._args, **new_kwargs)

        elif isinstance(field, ListSerializer) and isinstance(field.child, ModelSerializer):
            new_child = convert_serializer_fields_to_not_required(field.child.__class__, None, top_level=False)
            new_kwargs = field.child._kwargs | {"required": False, "many": True}
            new_class._declared_fields[field_name] = new_child(*field.child._args, **new_kwargs)

    return new_class


def _set_extra_kwargs(
    field_name: FieldNameStr,
    meta: SerializerMeta,
    lookup_field: FieldNameStr,
    *,
    top_level: bool,
) -> None:
    """Set `meta.extra_kwargs` settings for the given `field_name`."""
    # If the `field_name` is for a model property, should not set anything.
    attr = getattr(meta.model, field_name, None)
    if isinstance(attr, property) and field_name != lookup_field:
        return

    # If the `field_name` is the given `lookup field` while `top_level=True`, should set `required=True`,
    # otherwise should set `required=False`, unless the value has been explicitly set already.
    required = top_level and field_name == lookup_field
    field_options = meta.extra_kwargs.setdefault(field_name, {})
    field_options.setdefault("required", required)

    # Lookup field should be additionally marked as writeable so that
    # serializer doesn't remove it during validation.
    if field_name == lookup_field:
        meta.extra_kwargs.setdefault(field_name, {})["read_only"] = False



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from django.db import models
from rest_framework.exceptions import NotFound

if TYPE_CHECKING:
    from django.db.models import ForeignObjectRel
    from django.db.models.fields.related import RelatedField

    from .typing import Any, RelationType


__all__ = [
    "get_object_or_404",
    "get_model_lookup_field",
    "get_related_field_info",
]


def get_object_or_404(model_class: type[models.Model], **kwargs: Any) -> models.Model:
    try:
        return model_class._default_manager.get(**kwargs)
    except model_class.DoesNotExist as error:
        msg = f"`{model_class.__name__}` object matching query `{kwargs}` does not exist."
        raise NotFound(msg) from error


def get_model_lookup_field(model_class: type[models.Model], lookup_field: str | None = None) -> str:
    if lookup_field is None:
        # Use model primary key as lookup field.
        # This is usually the 'id' field, in which case we use 'pk' instead
        # to avoid collision with the 'id' field in GraphQL Relay nodes.
        lookup_field = model_class._meta.pk.name
        if lookup_field == "id":
            lookup_field = "pk"
    return lookup_field


@dataclass
class RelatedFieldInfo:
    """Information about a related field on a model."""

    name: str
    forward: bool
    relation: RelationType

    @property
    def one_to_one(self) -> bool:
        return self.relation == "one_to_one"

    @property
    def many_to_one(self) -> bool:
        return self.relation == "many_to_one"

    @property
    def one_to_many(self) -> bool:
        return self.relation == "one_to_many"

    @property
    def many_to_many(self) -> bool:
        return self.relation == "many_to_many"

    @property
    def reverse(self) -> bool:
        return not self.forward


def get_related_field_info(model: type[models.Model]) -> dict[str, RelatedFieldInfo]:
    """Map of all related fields on the given model to their related entity's field names."""
    mapping: dict[str, RelatedFieldInfo] = {}
    for field in model._meta.get_fields():
        if isinstance(field, (models.OneToOneRel, models.ManyToOneRel, models.ManyToManyRel)):
            name: str = field.get_accessor_name() or field.name
            mapping[name] = RelatedFieldInfo(
                name=field.remote_field.name,
                forward=False,
                relation=_get_relation_type(field),
            )

        if isinstance(field, (models.OneToOneField, models.ForeignKey, models.ManyToManyField)):
            value = field.remote_field.get_accessor_name() or field.name
            mapping[field.name] = RelatedFieldInfo(
                name=value,
                forward=True,
                relation=_get_relation_type(field),
            )
            continue

    return mapping


def _get_relation_type(field: ForeignObjectRel | RelatedField) -> RelationType:
    return (
        "one_to_one"
        if field.one_to_one
        else "many_to_one"
        if field.many_to_one
        else "one_to_many"
        if field.one_to_many
        else "many_to_many"
    )



from django.test.signals import setting_changed
from settings_holder import SettingsHolder, reload_settings

from .typing import Any, NamedTuple, Union

SETTING_NAME: str = "GRAPHENE_DJANGO_EXTENSIONS"


class DefaultSettings(NamedTuple):
    QUERY_PERMISSION_ERROR_MESSAGE: str = "No permission to access node."
    FILTER_PERMISSION_ERROR_MESSAGE: str = "No permission to access node."
    MUTATION_PERMISSION_ERROR_MESSAGE: str = "No permission to mutate."
    ORDERING_FILTER_NAME: str = "order_by"
    PERMISSION_DENIED_MESSAGE: str = "You do not have permission to access this field."


DEFAULTS: dict[str, Any] = DefaultSettings()._asdict()

IMPORT_STRINGS: set[Union[bytes, str]] = set()
REMOVED_SETTINGS: set[str] = set()

gdx_settings = SettingsHolder(
    setting_name=SETTING_NAME,
    defaults=DEFAULTS,
    import_strings=IMPORT_STRINGS,
    removed_settings=REMOVED_SETTINGS,
)

reload_my_settings = reload_settings(SETTING_NAME, gdx_settings)
setting_changed.connect(reload_my_settings)
